#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Enumerations for opcodes and status codes."""
__all__ = ["HTTPStatusCode", "GatewayCloseCode", "GatewayOpcode", "JSONErrorCode", "GatewayIntent"]

import enum


# Doesnt work correctly with enums, so since this file is all enums, ignore
# pylint: disable=no-member
@enum.unique
class HTTPStatusCode(enum.IntEnum):
    """HTTP response codes expected from RESTful components."""

    #: Continue
    CONTINUE = 100

    #: OK
    OK = 200

    #: Created
    CREATED = 201

    #: Accepted
    ACCEPTED = 202

    #: No content
    NO_CONTENT = 204

    #: Moved permanently
    MOVED_PERMANENTLY = 301

    #: Bad request
    BAD_REQUEST = 400

    #: Unauthorized
    UNAUTHORIZED = 401

    #: Forbidden
    FORBIDDEN = 403

    #: Not found
    NOT_FOUND = 404

    #: Method not allowed
    METHOD_NOT_ALLOWED = 405

    #: Not acceptable
    NOT_ACCEPTABLE = 406

    #: Proxy authentication required
    PROXY_AUTHENTICATION_REQUIRED = 407

    #: Request entitiy too large
    REQUEST_ENTITY_TOO_LARGE = 413

    #: Request URI too long
    REQUEST_URI_TOO_LONG = 414

    #: Unsupported media type
    UNSUPPORTED_MEDIA_TYPE = 415

    #: Im a teapot
    IM_A_TEAPOT = 418

    #: Too many requests
    TOO_MANY_REQUESTS = 429

    #: Internal server error
    INTERNAL_SERVER_ERROR = 500

    #: Not implemented
    NOT_IMPLEMENTED = 501

    #: Bad gateway
    BAD_GATEWAY = 502

    #: Service unavailable
    SERVICE_UNAVAILABLE = 503

    #: Gateway timeout
    GATEWAY_TIMEOUT = 504

    #: HTTP Version not supported
    HTTP_VERSION_NOT_SUPPORTED = 505

    def __str__(self) -> str:
        name = self.name.replace("_", " ").title() if self is not HTTPStatusCode.IM_A_TEAPOT else "I'm a teapot"
        return f"{self.value} {name}"


@enum.unique
class GatewayCloseCode(enum.IntEnum):
    """Reasons for closing a gateway connection.

    Note
    ----
    Any codes greater than or equal to ``4000`` are server-side codes. Any codes
    between ``1000`` and ``1999`` inclusive are generally client-side codes.
    """

    #: The application running closed.
    NORMAL_CLOSURE = 1000

    #: Discord is not sure what went wrong. Try reconnecting?
    UNKNOWN_ERROR = 4000

    #: You sent an invalid Gateway opcode or an invalid payload for an opcode.
    #: Don't do that!
    UNKNOWN_OPCODE = 4001

    #: You sent an invalid payload to Discord. Don't do that!
    DECODE_ERROR = 4002

    #: You sent Discord a payload prior to IDENTIFYing.
    NOT_AUTHENTICATED = 4003

    #: The account token sent with your identify payload is incorrect.
    AUTHENTICATION_FAILED = 4004

    #: You sent more than one identify payload. Don't do that!
    ALREADY_AUTHENTICATED = 4005

    #: The sequence sent when resuming the session was invalid. Reconnect and
    #: start a new session.
    INVALID_SEQ = 4007

    #: Woah nelly! You're sending payloads to Discord too quickly. Slow it down!
    RATE_LIMITED = 4008

    #: Your session timed out. Reconnect and start a new one.
    SESSION_TIMEOUT = 4009

    #: You sent Discord an invalid shard when IDENTIFYing.
    INVALID_SHARD = 4010

    #: The session would have handled too many guilds - you are required to
    #: shard your connection in order to connect.
    SHARDING_REQUIRED = 4011

    #: You sent an invalid version for the gateway.
    INVALID_VERSION = 4012

    #: You sent an invalid intent for a Gateway Intent. You may have incorrectly
    #: calculated the bitwise value.
    INVALID_INTENT = 4013

    #: You sent a disallowed intent for a Gateway Intent. You may have tried to
    #: specify an intent that you have not enabled or are not whitelisted for.
    DISALLOWED_INTENT = 4014

    def __str__(self) -> str:
        name = self.name.replace("_", " ").title()
        return f"{self.value} {name}"


@enum.unique
class GatewayOpcode(enum.IntEnum):
    """Opcodes that the gateway uses internally."""

    #: An event was dispatched.
    DISPATCH = 0

    #: Used for ping checking.
    HEARTBEAT = 1

    #: Used for client handshake.
    IDENTIFY = 2

    #: Used to update the client status.
    PRESENCE_UPDATE = 3

    #: Used to join/move/leave voice channels.
    VOICE_STATE_UPDATE = 4

    #: Used to resume a closed connection.
    RESUME = 6

    #: Used to tell clients to reconnect to the gateway.
    RECONNECT = 7

    #: Used to request guild members.
    REQUEST_GUILD_MEMBERS = 8

    #: Used to notify client they have an invalid session id.
    INVALID_SESSION = 9

    #: Sent immediately after connecting, contains heartbeat and server debug
    #: information.
    HELLO = 10

    #: Sent immediately following a client heartbeat that was received.
    HEARTBEAT_ACK = 11

    #: Not yet documented, so do not use.
    GUILD_SYNC = 12

    def __str__(self) -> str:
        name = self.name.replace("_", " ").title()
        return f"{self.value} {name}"


@enum.unique
class JSONErrorCode(enum.IntEnum):
    """Error codes that can be returned by the REST API."""

    #: This is sent if the payload is screwed up, etc.
    GENERAL_ERROR = 0

    #: Unknown account
    UNKNOWN_ACCOUNT = 10_001

    #: Unknown application
    UNKNOWN_APPLICATION = 10_002

    #: Unknown channel
    UNKNOWN_CHANNEL = 10_003

    #: Unknown guild
    UNKNOWN_GUILD = 10_004

    #: Unknown integration
    UNKNOWN_INTEGRATION = 10_005

    #: Unknown invite
    UNKNOWN_INVITE = 10_006

    #: Unknown member
    UNKNOWN_MEMBER = 10_007

    #: Unknown message
    UNKNOWN_MESSAGE = 10_008

    #: Unknown overwrite
    UNKNOWN_OVERWRITE = 10_009

    #: Unknown provider
    UNKNOWN_PROVIDER = 10_010

    #: Unknown role
    UNKNOWN_ROLE = 10_011

    #: Unknown token
    UNKNOWN_TOKEN = 10_012

    #: Unknown user
    UNKNOWN_USER = 10_013

    #: Unknown Emoji
    UNKNOWN_EMOJI = 10_014

    #: Unknown Webhook
    UNKNOWN_WEBHOOK = 10_015

    #: Bots cannot use this endpoint
    #:
    #: Note
    #: ----
    #: You should never expect to receive this in normal API usage.
    USERS_ONLY = 20_001

    #: Only bots can use this endpoint.
    #:
    #: Note
    #: ----
    #: You should never expect to receive this in normal API usage.
    BOTS_ONLY = 20_002

    #: Maximum number of guilds reached (100)
    #:
    #: Note
    #: ----
    #: You should never expect to receive this in normal API usage as this only
    #: applies to user accounts.
    #:
    #: This is unlimited for bot accounts.
    MAX_GUILDS_REACHED = 30_001

    #: Maximum number of friends reached (1000)
    #:
    #: Note
    #: ----
    #: You should never expect to receive this in normal API usage as this only
    #: applies to user accounts.
    #:
    #: Bots cannot have friends :( .
    MAX_FRIENDS_REACHED = 30_002

    #: Maximum number of pins reached (50)
    MAX_PINS_REACHED = 30_003

    #: Maximum number of guild roles reached (250)
    MAX_GUILD_ROLES_REACHED = 30_005

    #: Maximum number of webhooks reached (10)
    MAX_WEBHOOKS_REACHED = 30_007

    #: Maximum number of reactions reached (20)
    MAX_REACTIONS_REACHED = 30_010

    #: Maximum number of guild channels reached (500)
    MAX_GUILD_CHANNELS_REACHED = 30_013

    #: Maximun number of attachments in a message reached (10)
    MAX_MESSAGE_ATTACHMENTS_REACHED = 30_015

    #: Maximun number of invites reached (10000)
    MAX_INVITES_REACHED = 30_016

    #: Unauthorized
    UNAUTHORIZED = 40_001

    #: Request entity too large. Try sending something smaller in size
    TOO_LARGE = 40_005

    #: This feature has been temporarily disabled server-side
    DISABLED_TEMPORARILY = 40_006

    #: The user is banned from this guild
    USER_BANNED = 40_007

    #: Missing access
    MISSING_ACCESS = 50_001

    #: Invalid account type
    INVALID_ACCOUNT_TYPE = 50_002

    #: Cannot execute action on a DM channel
    CANNOT_EXECUTE_ACTION_ON_DM_CHANNEL = 50_003

    #: Widget Disabled
    WIDGET_DISABLED = 50_004

    #: Cannot edit a message authored by another user
    CANNOT_EDIT_A_MESSAGE_AUTHORED_BY_ANOTHER_USER = 50_005

    #: Cannot send an empty message
    CANNOT_SEND_AN_EMPTY_MESSAGE = 50_006

    #: Cannot send messages to this user
    CANNOT_SEND_MESSAGES_TO_THIS_USER = 50_007

    #: Cannot send messages in a voice channel
    CANNOT_SEND_MESSAGES_IN_VOICE_CHANNEL = 50_008

    #: Channel verification level is too high
    CHANNEL_VERIFICATION_TOO_HIGH = 50_009

    #: OAuth2 application does not have a bot
    OAUTH2_APPLICATION_DOES_NOT_HAVE_A_BOT = 50_010

    #: OAuth2 application limit reached
    OAUTH2_APPLICATION_LIMIT_REACHED = 50_011

    #: Invalid OAuth state
    INVALID_OAUTH2_STATE = 50_012

    #: Missing permissions
    MISSING_PERMISSIONS = 50_013

    #: Invalid authentication token
    INVALID_AUTHENTICATION_TOKEN = 50_014

    #: Note is too long
    NOTE_IS_TOO_LONG = 50_015

    #: Provided too few or too many messages to delete. Must provide at least 2
    #: and fewer than 100 messages to delete.
    INVALID_NUMBER_OF_MESSAGES_TO_DELETE = 50_016

    #: A message can only be pinned to the channel it was sent in
    CANNOT_PIN_A_MESSAGE_IN_A_DIFFERENT_CHANNEL = 50_019

    #: Invite code is either invalid or taken.
    INVALID_INVITE = 50_020

    #: Cannot execute action on a system message
    CANNOT_EXECUTE_ACTION_ON_SYSTEM_MESSAGE = 50_021

    #: Invalid OAuth2 access token
    INVALID_OAUTH2_TOKEN = 50_025

    #: A message provided was too old to bulk delete
    MESSAGE_PROVIDED_WAS_TOO_OLD_TO_BULK_DELETE = 50_034

    #: Invalid Form Body
    INVALID_FORM_BODY = 50_035

    #: An invite was accepted to a guild the application's bot is not in
    ACCEPTED_INVITE_TO_GUILD_BOT_IS_NOT_IN = 50_036

    #: Invalid API version
    INVALID_API_VERSION = 50_041

    #: Reaction blocked
    REACTION_BLOCKED = 90_001

    #: The resource is overloaded.
    RESOURCE_OVERLOADED = 130_000

    def __str__(self) -> str:
        name = self.name.replace("_", " ").title()
        return f"{self.value} {name}"


@enum.unique
class GatewayIntent(enum.IntFlag):
    """Represents an intent on the gateway.

    This is a bitfield representation of all the categories of event
    that you wish to receive.

    Any events not in an intent category will be fired regardless of what
    intents you provide.

    Warnings
    --------
    If you are using the V7 Gateway, you will be REQUIRED to provide some form
    of intent value when you connect. Failure to do so may result in immediate
    termination of the session server-side.

    Notes
    -----
    Discord now places limits on certain events you can receive without
    whitelisting your bot first. On the ``Bot`` tab in the developer's portal
    for your bot, you should now have the option to enable functionality
    for receiving these events.

    If you attempt to request an intent type that you have not whitelisted
    your bot for, you will be disconnected on startup with a ``4014`` closure
    code.
    """

    #: Subscribes to the following events:
    #: * GUILD_CREATE
    #: * GUILD_UPDATE
    #: * GUILD_DELETE
    #: * GUILD_ROLE_CREATE
    #: * GUILD_ROLE_UPDATE
    #: * GUILD_ROLE_DELETE
    #: * CHANNEL_CREATE
    #: * CHANNEL_UPDATE
    #: * CHANNEL_DELETE
    #: * CHANNEL_PINS_UPDATE
    GUILDS = 1 << 0

    #: Subscribes to the following events:
    #: * GUILD_MEMBER_ADD
    #: * GUILD_MEMBER_UPDATE
    #: * GUILD_MEMBER_REMOVE
    #:
    #: Warnings
    #: --------
    #: This intent is privileged, and requires enabling/whitelisting to use.
    GUILD_MEMBERS = 1 << 1

    #: Subscribes to the following events:
    #: * GUILD_BAN_ADD
    #: * GUILD_BAN_REMOVE
    GUILD_BANS = 1 << 2

    #: Subscribes to the following events:
    #: * GUILD_EMOJIS_UPDATE
    GUILD_EMOJIS = 1 << 3

    #: Subscribes to the following events:
    #: * GUILD_INTEGRATIONS_UPDATE
    GUILD_INTEGRATIONS = 1 << 4

    #: Subscribes to the following events:
    #: * WEBHOOKS_UPDATE
    GUILD_WEBHOOKS = 1 << 5

    #: Subscribes to the following events:
    #: * INVITE_CREATE
    #: * INVITE_DELETE
    GUILD_INVITES = 1 << 6

    #: Subscribes to the following events:
    #: * VOICE_STATE_UPDATE
    GUILD_VOICE_STATES = 1 << 7

    #: Subscribes to the following events:
    #: * PRESENCE_UPDATE
    #:
    #: Warnings
    #: --------
    #: This intent is privileged, and requires enabling/whitelisting to use.
    GUILD_PRESENCES = 1 << 8

    #: Subscribes to the following events:
    #: * MESSAGE_CREATE
    #: * MESSAGE_UPDATE
    #: * MESSAGE_DELETE
    #: * MESSAGE_BULK
    GUILD_MESSAGES = 1 << 9

    #: Subscribes to the following events:
    #: * MESSAGE_REACTION_ADD
    #: * MESSAGE_REACTION_REMOVE
    #: * MESSAGE_REACTION_REMOVE_ALL
    #: * MESSAGE_REACTION_REMOVE_EMOJI
    GUILD_MESSAGE_REACTIONS = 1 << 10

    #: Subscribes to the following events:
    #: * TYPING_START
    GUILD_MESSAGE_TYPING = 1 << 11

    #: Subscribes to the following events:
    #: * CHANNEL_CREATE
    #: * MESSAGE_CREATE
    #: * MESSAGE_UPDATE
    #: * MESSAGE_DELETE
    DIRECT_MESSAGES = 1 << 12

    #: Subscribes to the following events:
    #: * MESSAGE_REACTION_ADD
    #: * MESSAGE_REACTION_REMOVE
    #: * MESSAGE_REACTION_REMOVE_ALL
    DIRECT_MESSAGE_REACTIONS = 1 << 13

    #: Subscribes to the following events
    #: * TYPING_START
    DIRECT_MESSAGE_TYPING = 1 << 14


# pylint: enable=no-member



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""An executable module to be used to test that the gateway works as intended.

This is only for use by developers of this library, regular users do not need
to use this.
"""
import logging
import os
import sys
import typing

import click

from hikari.clients import configs
from hikari.clients import gateway_managers
from hikari.state import stateless_event_managers


_LOGGER_LEVELS: typing.Final[typing.Sequence[str]] = ["DEBUG", "INFO", "WARNING", "ERROR", "NOTSET"]


def _supports_color():
    plat = sys.platform
    supported_platform = plat != "Pocket PC" and (plat != "win32" or "ANSICON" in os.environ)
    # isatty is not always implemented, #6223.
    is_a_tty = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()
    return supported_platform and is_a_tty


_COLOR_FORMAT: typing.Final[str] = (
    "\033[1;35m%(levelname)1.1s \033[0;37m%(name)25.25s \033[0;31m%(asctime)23.23s \033[1;34m%(module)-15.15s "
    "\033[1;32m#%(lineno)-4d \033[0m:: \033[0;33m%(message)s\033[0m"
)

_REGULAR_FORMAT: typing.Final[str] = (
    "%(levelname)1.1s %(name)25.25s %(asctime)23.23s %(module)-15.15s #%(lineno)-4d :: %(message)s"
)


@click.command()
@click.option("--compression", default=True, type=click.BOOL, help="Enable or disable gateway compression.")
@click.option("--color", default=_supports_color(), type=click.BOOL, help="Whether to enable or disable color.")
@click.option("--debug", default=False, type=click.BOOL, help="Enable or disable debug mode.")
@click.option("--logger", envvar="LOGGER", default="INFO", type=click.Choice(_LOGGER_LEVELS), help="Logger verbosity.")
@click.option("--shards", default=1, type=click.IntRange(min=1), help="The number of shards to explicitly use.")
@click.option("--token", required=True, envvar="TOKEN", help="The token to use to authenticate with Discord.")
@click.option("--url", default="wss://gateway.discord.gg/", help="The websocket URL to connect to.")
@click.option("--verify-ssl", default=True, type=click.BOOL, help="Enable or disable SSL verification.")
@click.option("--version", default=6, type=click.IntRange(min=6), help="Version of the gateway to use.")
def run_gateway(compression, color, debug, logger, shards, token, url, verify_ssl, version) -> None:
    """:mod:`click` command line client for running a test gateway connection.

    This is provided for internal testing purposes for benchmarking API
    stability, etc.
    """
    logging.captureWarnings(True)

    logging.basicConfig(level=logger, format=_COLOR_FORMAT if color else _REGULAR_FORMAT, stream=sys.stdout)

    manager = stateless_event_managers.StatelessEventManagerImpl()

    client = gateway_managers.GatewayManager(
        shard_ids=[*range(shards)],
        shard_count=shards,
        config=configs.WebsocketConfig(
            token=token,
            gateway_version=version,
            debug=debug,
            gateway_use_compression=compression,
            verify_ssl=verify_ssl,
        ),
        url=url,
        raw_event_consumer_impl=manager,
        dispatcher=manager.event_dispatcher,
    )

    client.run()


if __name__ == "__main__":
    run_gateway()  # pylint:disable=no-value-for-parameter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Complex rate limiting mechanisms.

Provides an implementation for the complex rate limiting mechanisms that Discord
requires for rate limit handling that conforms to the passed bucket headers
correctly.

What is the theory behind this implementation?
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

In this module, we refer to a :obj:`hikari.net.routes.CompiledRoute` as a
definition of a route with specific major parameter values included (e.g.
``POST /channels/123/messages``), and a :obj:`hikari.net.routes.RouteTemplate`
as a definition of a route without specific parameter values included (e.g.
``POST /channels/{channel_id}/messages``). We can compile a
:obj:`hikari.net.routes.CompiledRoute` from a
:obj:`hikari.net.routes.RouteTemplate` by providing the corresponding
parameters as kwargs, as you may already know.

In this module, a "bucket" is an internal data structure that tracks and
enforces the  rate limit state for a specific
:obj:`hikari.net.routes.CompiledRoute`,  and can manage delaying tasks in the
event that we begin to get rate limited. It also supports providing in-order
execution of queued tasks.

Discord allocates types of buckets to routes. If you are making a request and
there is a valid rate limit on the route you hit, you should receive an
``X-RateLimit-Bucket`` header from the server in your response. This is a hash
that identifies a route based on internal criteria that does not include major
parameters. This ``X-RateLimitBucket`` is known in this module as an
"bucket hash".

This means that generally, the route `POST /channels/123/messages` and
``POST /channels/456/messages`` will usually sit in the same bucket, but
``GET /channels/123/messages/789`` and ``PATCH /channels/123/messages/789`` will
usually not share the same bucket. Discord may or may not change this at any
time, so hard coding this logic is not a useful thing to be doing.

Rate limits, on the other hand, apply to a bucket and are specific to the major
parameters of the compiled route. This means that ``POST /channels/123/messages``
and ``POST /channels/456/messages`` do not share the same real bucket, despite
Discord providing the same bucket hash. A real bucket hash is the :obj:`str`
hash of the bucket that Discord sends us in a response concatenated to the
corresponding major parameters. This is used for quick bucket indexing
internally in this module.

One issue that occurs from this is that we cannot effectively hash a
:obj:`hikari.net.routes.CompiledRoute` that has not yet been hit, meaning that
until we receive a response from this endpoint, we have no idea what our rate
limits could be, nor the bucket that they sit in. This is usually not
problematic, as the first request to an endpoint should never be rate limited
unless you are hitting it from elsewhere in the same time window outside your
Hikari application. To manage this situation, unknown endpoints are allocated to
a special unlimited bucket until they have an initial bucket hash code allocated
from a response. Once this happens, the route is reallocated a dedicated bucket.
Unknown buckets have a hardcoded initial hash code internally.

Initially acquiring time on a bucket
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Each time you ``BaseRateLimiter.acquire()`` a request
timeslice for a given :obj:`hikari.net.routes.CompiledRoute`, several
things happen. The first is that we attempt to find the existing bucket for
that route, if there is one, or get an unknown bucket otherwise. This is done
by creating a real bucket hash` from the compiled route. The initial hash
is calculated using a lookup table that maps :obj:`hikari.net.routes.CompiledRoute`
objects to their corresponding initial hash codes, or to the unknown bucket
hash code if not yet known. This initial hash is processed by the
:obj:`hikari.net.routes.CompiledRoute` to provide the real bucket hash we
need to get the route's bucket object internally.

The ``acquire`` method will take the bucket and acquire a new timeslice on
it. This takes the form of a :obj:`asyncio.Future` which should be awaited by
the caller and will complete once the caller is allowed to make a request. Most
of the time, this is done instantly, but if the bucket has an active rate limit
preventing requests being sent, then the future will be paused until the rate
limit is over. This may be longer than the rate limit period if you have queued
a large number of requests during this limit, as it is first-come-first-served.

Acquiring a rate limited bucket will start a bucket-wide task (if not already
running) that will wait until the rate limit has completed before allowing more
futures to complete. This is done while observing the rate limits again, so can
easily begin to re-ratelimit itself if needed. Once the task is complete, it
tidies itself up and disposes of itself. This task will complete once the queue
becomes empty.

The result of ``RateLimiter.acquire()`` is a tuple of a
:obj:`asyncio.Future` to await on which completes when you are allowed to
proceed with making a request, and a real bucket hash which should be
stored temporarily. This will be explained in the next section.

When you make your response, you should be sure to set the
``X-RateLimit-Precision`` header to ``millisecond`` to ensure a much greater
accuracy against rounding errors for rate limits (reduces the error margin from
``1`` second to ``1`` millisecond).

Handling the rate limit headers of a response
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Once you have received your response, you are expected to extract the values of
the vital rate limit headers manually and parse them to the correct data types.
These headers are:

* ``Date``:
    the response date on the server. This should be parsed to a
    :obj:`datetime.datetime` using :func:`email.utils.parsedate_to_datetime`.
* ``X-RateLimit-Limit``:
    an :obj:`int` describing the max requests in the bucket
    from empty to being rate limited.
* ``X-RateLimit-Remaining``:
    an :obj:`int` describing the remaining number of
    requests before rate limiting occurs in the current window.
* ``X-RateLimit-Bucket``:
    a :obj:`str` containing the initial bucket hash.
* ``X-RateLimit-Reset``:
    a :obj:`float` containing the number of seconds since
    1st January 1970 at 0:00:00 UTC at which the current ratelimit window
    resets. This should be parsed to a :obj:`datetime.datetime` using
    :meth:`datetime.datetime.fromtimestamp`, passing :obj:`datetime.timezone.utc`
    as ``tz``.

Each of the above values should be passed to the
``update_rate_limits`` method to ensure that the bucket you acquired time
from is correctly updated should Discord decide to alter their ratelimits on the
fly without warning (including timings and the bucket).

This method will manage creating new buckets as needed and resetting vital
information in each bucket you use.

Tidying up
~~~~~~~~~~

To prevent unused buckets cluttering up memory, each :obj:`BaseRateLimiter`
instance spins up a :obj:`asyncio.Task` that periodically locks the bucket
list (not threadsafe, only using the concept of asyncio not yielding in regular
functions) and disposes of any clearly stale buckets that are no longer needed.
These will be recreated again in the future if they are needed.

When shutting down an application, one must remember to ``close()`` the
:obj:`BaseRateLimiter` that has been used. This will ensure the garbage collection
task is stopped, and will also ensure any remaining futures in any bucket queues
have an :obj:`asyncio.CancelledError` set on them to prevent deadlocking
ratelimited calls that may be waiting to be unlocked.
"""
__all__ = [
    "BaseRateLimiter",
    "BurstRateLimiter",
    "ManualRateLimiter",
    "WindowedBurstRateLimiter",
    "HTTPBucketRateLimiter",
    "HTTPBucketRateLimiterManager",
    "ExponentialBackOff",
]

import abc
import asyncio
import datetime
import logging
import random
import time
import types
import typing
import weakref

from hikari.internal import more_asyncio
from hikari.internal import more_logging
from hikari.net import routes

#: The hash used for an unknown bucket that has not yet been resolved.
#:
#: :type: :obj:`str`
UNKNOWN_HASH: typing.Final[str] = "UNKNOWN"


class BaseRateLimiter(abc.ABC):
    """Base for any asyncio-based rate limiter being used.

    Supports being used as a synchronous context manager.

    Warnings
    --------
    Async context manager support is not supported and will not be supported.
    """

    __slots__ = ()

    @abc.abstractmethod
    def acquire(self) -> more_asyncio.Future[None]:
        """Acquire permission to perform a task that needs to have rate limit management enforced.

        Returns
        -------
        :obj:`asyncio.Future`
            A future that should be awaited. Once the future is complete, you
            can proceed to execute your rate-limited task.
        """

    @abc.abstractmethod
    def close(self) -> None:
        """Close the rate limiter, cancelling any internal tasks that are executing."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class BurstRateLimiter(BaseRateLimiter, abc.ABC):
    """Base implementation for a burst-based rate limiter.

    This provides an internal queue and throttling placeholder, as well as
    complete logic for safely aborting any pending tasks when being shut down.
    """

    __slots__ = ("name", "throttle_task", "queue", "logger")

    #: The name of the rate limiter.
    #:
    #: :type: :obj:`str`
    name: typing.Final[str]

    #: The throttling task, or :obj:`None` if it isn't running.
    #:
    #: :type: :obj:`asyncio.Task`, optional
    throttle_task: typing.Optional[more_asyncio.Task[None]]

    #: The queue of any futures under a rate limit.
    #:
    #: :type: :obj:`asyncio.Queue` [`asyncio.Future`]
    queue: typing.Final[typing.List[more_asyncio.Future[None]]]

    #: The logger used by this rate limiter.
    #:
    #: :type: :obj:`logging.Logger`
    logger: typing.Final[logging.Logger]

    def __init__(self, name: str) -> None:
        self.name = name
        self.throttle_task = None
        self.queue = []
        self.logger: logging.Logger = more_logging.get_named_logger(self)

    @abc.abstractmethod
    def acquire(self) -> more_asyncio.Future[None]:
        """Acquire time on this rate limiter.

        The implementation should define this.

        Returns
        -------
        :obj:`asyncio.Future`
            A future that should be immediately awaited. Once the await
            completes, you are able to proceed with the operation that is
            under this rate limit.
        """

    def close(self) -> None:
        """Close the rate limiter, and shut down any pending tasks.

        Once this is invoked, you should not reuse this object.
        """
        if self.throttle_task is not None:
            self.throttle_task.cancel()

        failed_tasks = 0
        while self.queue:
            failed_tasks += 1
            future = self.queue.pop(0)
            # Make the future complete with an exception
            future.cancel()

        if failed_tasks:
            self.logger.error("%s rate limiter closed with %s pending tasks!", self.name, failed_tasks)
        else:
            self.logger.debug("%s rate limiter closed", self.name)

    @property
    def is_empty(self) -> bool:
        """Return :obj:`True` if no futures are on the queue being rate limited."""
        return len(self.queue) == 0


class ManualRateLimiter(BurstRateLimiter):
    """Rate limit handler for the global HTTP rate limit.

    This is a non-preemptive rate limiting algorithm that will always return
    completed futures until :meth:`throttle` is invoked. Once this is invoked,
    any subsequent calls to :meth:`acquire` will return incomplete futures
    that will be enqueued to an internal queue. A task will be spun up to wait
    for a period of time given to the :meth:`throttle`. Once that has passed,
    the lock will begin to re-consume incomplete futures on the queue,
    completing them.

    Triggering a throttle when it is already set will cancel the current
    throttle task that is sleeping and replace it.

    This is used to enforce the global HTTP rate limit that will occur
    "randomly" during HTTP API interaction.

    Expect random occurrences.
    """

    __slots__ = ()

    def __init__(self) -> None:
        super().__init__("global HTTP")

    def acquire(self) -> more_asyncio.Future[None]:
        """Acquire time on this rate limiter.

        Returns
        -------
        :obj:`asyncio.Future`
            A future that should be immediately awaited. Once the await
            completes, you are able to proceed with the operation that is
            under this rate limit.
        """
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        if self.throttle_task is not None:
            self.queue.append(future)
        else:
            future.set_result(None)
        return future

    def throttle(self, retry_after: float) -> None:
        """Perform the throttling rate limiter logic.

        Iterates repeatedly while the queue is not empty, adhering to any
        rate limits that occur in the mean time.

        retry_after : :obj:`float`
            How long to sleep for before unlocking and releasing any futures
            in the queue.

        Note
        ----
        This will invoke :meth:`unlock_later` as a scheduled task in the future
        (it will not await it to finish)

        When the :meth:`unlock_later` coroutine function completes, it should be
        expected to set the `throttle_task`` to :obj:`None`. This means you can
        check if throttling is occurring by checking if ``throttle_task``
        is not :obj:`None`.

        If this is invoked while another throttle is in progress, that one is
        cancelled and a new one is started. This enables new rate limits to
        override existing ones.
        """
        if self.throttle_task is not None:
            self.throttle_task.cancel()

        loop = asyncio.get_running_loop()
        self.throttle_task = loop.create_task(self.unlock_later(retry_after))

    async def unlock_later(self, retry_after: float) -> None:
        """Sleeps for a while, then removes the lock.

        Parameters
        ----------
        retry_after : :obj:`float`
            How long to sleep for before unlocking and releasing any futures
            in the queue.

        Note
        ----
        You shouldn't need to invoke this directly. Call :meth:`throttle`
        instead.

        When the :meth:`unlock_later` coroutine function completes, it should be
        expected to set the ``throttle_task`` to :obj:`None`. This means you can
        check if throttling is occurring by checking if ``throttle_task``
        is not :obj:`None`.
        """
        self.logger.warning("you are being globally rate limited for %ss", retry_after)
        await asyncio.sleep(retry_after)
        while self.queue:
            next_future = self.queue.pop(0)
            next_future.set_result(None)
        self.throttle_task = None


class WindowedBurstRateLimiter(BurstRateLimiter):
    """Windowed burst rate limiter.

    Rate limiter for rate limits that last fixed periods of time with a
    fixed number of times it can be used in that time frame.

    To use this, you should call :meth:`acquire` and await the result
    immediately before performing your rate-limited task.

    If the rate limit has been hit, acquiring time will return an incomplete
    future that is placed on the internal queue. A throttle task is then spun up
    if not already running that will be expected to provide some implementation
    of backing off and sleeping for a given period of time until the limit has
    passed, and then proceed to consume futures from the queue while adhering
    to those rate limits.

    If the throttle task is already running, the acquired future will always be
    incomplete and enqueued regardless of whether the rate limit is actively
    reached or not.

    Acquiring a future from this limiter when no throttling task is running and
    when the rate limit is not reached will always result in the task invoking
    a drip and a completed future being returned.

    Dripping is left to the implementation of this class, but will be expected
    to provide some mechanism for updating the internal statistics to represent
    that a unit has been placed into the bucket.
    """

    __slots__ = ("reset_at", "remaining", "limit", "period")

    #: The :func:`time.perf_counter` that the limit window ends at.
    #:
    #: :type: :obj:`float`
    reset_at: float

    #: The number of :meth:`acquire`'s left in this window before you will get
    #: rate limited.
    #:
    #: :type: :obj:`int`
    remaining: int

    #: How long the window lasts for from the start in seconds.
    #:
    #: :type: :obj:`float`
    period: float

    #: The maximum number of :meth:`acquire`'s allowed in this time window.
    #:
    #: :type: :obj:`int`
    limit: int

    def __init__(self, name: str, period: float, limit: int) -> None:
        super().__init__(name)
        self.reset_at = 0.0
        self.remaining = 0
        self.limit = limit
        self.period = period

    def acquire(self) -> more_asyncio.Future[None]:
        """Acquire time on this rate limiter.

        Returns
        -------
        :obj:`asyncio.Future`
            A future that should be immediately awaited. Once the await
            completes, you are able to proceed with the operation that is
            under this rate limit.
        """
        loop = asyncio.get_running_loop()
        future = loop.create_future()

        # If we are rate limited, delegate invoking this to the throttler and spin it up
        # if it hasn't started. Likewise, if the throttle task is still running, we should
        # delegate releasing the future to the throttler task so that we still process
        # first-come-first-serve
        if self.throttle_task is not None or self.is_rate_limited(time.perf_counter()):
            self.queue.append(future)
            if self.throttle_task is None:
                self.throttle_task = loop.create_task(self.throttle())
        else:
            self.drip()
            future.set_result(None)

        return future

    def get_time_until_reset(self, now: float) -> float:
        """Determine how long until the current rate limit is reset.

        Parameters
        ----------
        now : :obj:`float`
            The monotonic :func:`time.perf_counter` timestamp.

        Returns
        -------
        :obj:`float`
            The time left to sleep before the rate limit is reset. If no rate limit
            is in effect, then this will return ``0.0`` instead.

        Warning
        -------
        Invoking this method will update the internal state if we were
        previously rate limited, but at the given time are no longer under that
        limit. This makes it imperative that you only pass the current timestamp
        to this function, and not past or future timestamps. The effects of
        doing the latter are undefined behaviour.
        """
        if not self.is_rate_limited(now):
            return 0.0
        return self.reset_at - now

    def is_rate_limited(self, now: float) -> bool:
        """Determine if we are under a rate limit at the given time.

        Parameters
        ----------
        now : :obj:`float`
            The monotonic :func:`time.perf_counter` timestamp.

        Returns
        -------
        :obj:`bool`
            :obj:`True` if we are being rate limited. :obj:`False` if we are not.

        Warning
        -------
        Invoking this method will update the internal state if we were
        previously rate limited, but at the given time are no longer under that
        limit. This makes it imperative that you only pass the current timestamp
        to this function, and not past or future timestamps. The effects of
        doing the latter are undefined behaviour.
        """
        if self.reset_at <= now:
            self.remaining = self.limit
            self.reset_at = now + self.period
            return False

        return self.remaining <= 0

    def drip(self) -> None:
        """Decrements the remaining counter."""
        self.remaining -= 1

    async def throttle(self) -> None:
        """Perform the throttling rate limiter logic.

        Iterates repeatedly while the queue is not empty, adhering to any
        rate limits that occur in the mean time.

        Note
        ----
        You should usually not need to invoke this directly, but if you do,
        ensure to call it using :func:`asyncio.create_task`, and store the
        task immediately in ``throttle_task``.

        When this coroutine function completes, it will set the
        ``throttle_task`` to :obj:`None`. This means you can check if throttling
        is occurring by checking if ``throttle_task`` is not :obj:`None`.
        """
        self.logger.debug(
            "you are being rate limited on bucket %s, backing off for %ss",
            self.name,
            self.get_time_until_reset(time.perf_counter()),
        )

        while self.queue:
            sleep_for = self.get_time_until_reset(time.perf_counter())
            await asyncio.sleep(sleep_for)

            while self.remaining > 0 and self.queue:
                self.drip()
                self.queue.pop(0).set_result(None)

        self.throttle_task = None


class HTTPBucketRateLimiter(WindowedBurstRateLimiter):
    """Represents a rate limit for an HTTP endpoint.

    Component to represent an active rate limit bucket on a specific HTTP route
    with a specific major parameter combo.

    This is somewhat similar to the :obj:`WindowedBurstRateLimiter` in how it
    works.

    This algorithm will use fixed-period time windows that have a given limit
    (capacity). Each time a task requests processing time, it will drip another
    unit into the bucket. Once the bucket has reached its limit, nothing can
    drip and new tasks will be queued until the time window finishes.

    Once the time window finishes, the bucket will empty, returning the current
    capacity to zero, and tasks that are queued will start being able to drip
    again.

    Additional logic is provided by the :meth:`update_rate_limit` call which
    allows dynamically changing the enforced rate limits at any time.
    """

    __slots__ = ("compiled_route",)

    #: The compiled route that this rate limit is covering.
    #:
    #: :type: :obj:`hikari.net.routes.CompiledRoute`
    compiled_route: typing.Final[routes.CompiledRoute]

    def __init__(self, name: str, compiled_route: routes.CompiledRoute) -> None:
        super().__init__(name, 1, 1)
        # We store this since the compiled route mapping acts as a weak key dictionary to aid in auto garbage
        # collecting itself; this acts as our solid reference.
        self.compiled_route = compiled_route

    @property
    def is_unknown(self) -> bool:
        """Return :obj:`True` if the bucket represents an ``UNKNOWN`` bucket."""
        return self.name.startswith(UNKNOWN_HASH)

    def acquire(self) -> more_asyncio.Future[None]:
        """Acquire time on this rate limiter.

        Returns
        -------
        :obj:`asyncio.Future`
            A future that should be awaited immediately. Once the future completes,
            you are allowed to proceed with your operation.

        Note
        ----
        You should afterwards invoke :meth:`update_rate_limit` to update any
        rate limit information you are made aware of.
        """
        return more_asyncio.completed_future(None) if self.is_unknown else super().acquire()

    def update_rate_limit(self, remaining: int, limit: int, reset_at: float) -> None:
        """Amend the rate limit.

        Parameters
        ----------
        remaining : :obj:`int`
            The calls remaining in this time window.
        limit : :obj:`int`
            The total calls allowed in this time window.
        reset_at : :obj:`float`
            The epoch at which to reset the limit.

        Note
        ----
        The ``reset_at`` epoch is expected to be a :func:`time.perf_counter`
        monotonic epoch, rather than a :func:`time.time` date-based epoch.
        """
        self.remaining = remaining
        self.limit = limit
        self.reset_at = reset_at
        self.period = max(0.0, self.reset_at - time.perf_counter())

    def drip(self) -> None:
        """Decrement the remaining count for this bucket.

        Note
        ----
        If the bucket is marked as :attr:`is_unknown`, then this will not do
        anything. ``Unknown`` buckets have infinite rate limits.
        """
        # We don't drip unknown buckets: we can't rate limit them as we don't know their real bucket hash or
        # the current rate limit values Discord put on them...
        if not self.is_unknown:
            self.remaining -= 1


class HTTPBucketRateLimiterManager:
    """The main rate limiter implementation for HTTP clients.

    This is designed to provide bucketed rate limiting for Discord HTTP
    endpoints that respects the ``X-RateLimit-Bucket`` rate limit header. To do
    this, it makes the assumption that any limit can change at any time.
    """

    __slots__ = (
        "routes_to_hashes",
        "real_hashes_to_buckets",
        "closed_event",
        "gc_task",
        "logger",
    )

    #: Maps compiled routes to their ``X-RateLimit-Bucket`` header being used.
    #:
    #: :type: :obj:`typing.MutableMapping` [ :obj:`hikari.net.routes.CompiledRoute`, :obj:`str` ]
    routes_to_hashes: typing.Final[typing.MutableMapping[routes.CompiledRoute, str]]

    #: Maps full bucket hashes (``X-RateLimit-Bucket`` appended with a hash of
    #: major parameters used in that compiled route) to their corresponding rate
    #: limiters.
    #:
    #: :type: :obj:`typing.MutableMapping` [ :obj:`str`, :obj:`HTTPBucketRateLimiter` ]
    real_hashes_to_buckets: typing.Final[typing.MutableMapping[str, HTTPBucketRateLimiter]]

    #: An internal event that is set when the object is shut down.
    #:
    #: :type: :obj:`asyncio.Event`
    closed_event: typing.Final[asyncio.Event]

    #: The internal garbage collector task.
    #:
    #: :type: :obj:`asyncio.Task`, optional
    gc_task: typing.Optional[more_asyncio.Task[None]]

    #: The logger to use for this object.
    #:
    #: :type: :obj:`logging.Logger`
    logger: typing.Final[logging.Logger]

    def __init__(self) -> None:
        self.routes_to_hashes = weakref.WeakKeyDictionary()
        self.real_hashes_to_buckets = {}
        self.closed_event: asyncio.Event = asyncio.Event()
        self.gc_task: typing.Optional[asyncio.Task] = None
        self.logger: logging.Logger = more_logging.get_named_logger(self)

    def __enter__(self) -> "HTTPBucketRateLimiterManager":
        return self

    def __exit__(self, exc_type: typing.Type[Exception], exc_val: Exception, exc_tb: types.TracebackType) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()

    def start(self, poll_period: float = 20) -> None:
        """Start this ratelimiter up.

        This spins up internal garbage collection logic in the background to
        keep memory usage to an optimal level as old routes and bucket hashes
        get discarded and replaced.

        Parameters
        ----------
        poll_period : :obj:`float`
            Period to poll the garbage collector at in seconds. Defaults
            to ``20`` seconds.
        """
        if not self.gc_task:
            self.gc_task = asyncio.get_running_loop().create_task(self.gc(poll_period))

    def close(self) -> None:
        """Close the garbage collector and kill any tasks waiting on ratelimits.

        Once this has been called, this object is considered to be effectively
        dead. To reuse it, one should create a new instance.
        """
        self.closed_event.set()
        for bucket in self.real_hashes_to_buckets.values():
            bucket.close()
        self.real_hashes_to_buckets.clear()
        self.routes_to_hashes.clear()

    # Ignore docstring not starting in an imperative mood
    async def gc(self, poll_period: float = 20) -> None:  # noqa: D401
        """The garbage collector loop.

        This is designed to run in the background and manage removing unused
        route references from the rate-limiter collection to save memory.

        This will run forever until :attr:`closed_event` is set. This will
        invoke :meth:`do_gc_pass` periodically.

        Parameters
        ----------
        poll_period : :obj:`float`
            The period to poll at. This defaults to once every ``20`` seconds.

        Warnings
        --------
        You generally have no need to invoke this directly. Use
        :meth:`start` and :meth:`close` to control this instead.
        """
        # Prevent filling memory increasingly until we run out by removing dead buckets every 20s
        # Allocations are somewhat cheap if we only do them every so-many seconds, after all.
        self.logger.debug("rate limit garbage collector started")
        while not self.closed_event.is_set():
            try:
                await asyncio.wait_for(self.closed_event.wait(), timeout=poll_period)
            except asyncio.TimeoutError:
                self.logger.debug("performing rate limit garbage collection pass")
                self.do_gc_pass()
        self.gc_task = None

    def do_gc_pass(self) -> None:
        """Perform a single garbage collection pass.

        This will assess any routes stored in the internal mappings of this
        object and remove any that are deemed to be inactive or dead in order
        to save memory.

        If the removed routes are used again in the future, they will be
        re-cached automatically.

        Warning
        -------
        You generally have no need to invoke this directly. Use
        :meth:`start` and :meth:`close` to control this instead.
        """
        buckets_to_purge = []

        # Discover and purge
        for full_hash, bucket in self.real_hashes_to_buckets.items():
            if bucket.is_empty and (bucket.is_unknown or bucket.reset_at < time.perf_counter()):
                # If it is still running a throttle and is in memory, it will remain in memory
                # but we won't know about it.
                buckets_to_purge.append(full_hash)

        for full_hash in buckets_to_purge:
            self.real_hashes_to_buckets[full_hash].close()
            del self.real_hashes_to_buckets[full_hash]

        self.logger.debug("purged %s stale buckets", len(buckets_to_purge))

    def acquire(self, compiled_route: routes.CompiledRoute) -> more_asyncio.Future:
        """Acquire a bucket for the given route.

        Parameters
        ----------
        compiled_route : :obj:`hikari.net.routes.CompiledRoute`
            The route to get the bucket for.

        Returns
        -------
        :obj:`asyncio.Future`
            A future to await that completes when you are allowed to run
            your request logic.

        Note
        ----
        The returned future MUST be awaited, and will complete when your turn to
        make a call comes along. You are expected to await this and then
        immediately make your HTTP call. The returned future may already be
        completed if you can make the call immediately.
        """
        # Returns a future to await on to wait to be allowed to send the request, and a
        # bucket hash to use to update rate limits later.
        if compiled_route in self.routes_to_hashes:
            bucket_hash = self.routes_to_hashes[compiled_route]
        else:
            bucket_hash = UNKNOWN_HASH
            self.routes_to_hashes[compiled_route] = bucket_hash

        real_bucket_hash = compiled_route.create_real_bucket_hash(bucket_hash)

        try:
            bucket = self.real_hashes_to_buckets[real_bucket_hash]
        except KeyError:
            self.logger.debug("creating new bucket for %s", real_bucket_hash)
            bucket = HTTPBucketRateLimiter(real_bucket_hash, compiled_route)
            self.real_hashes_to_buckets[real_bucket_hash] = bucket

        return bucket.acquire()

    def update_rate_limits(
        self,
        compiled_route: routes.CompiledRoute,
        bucket_header: typing.Optional[str],
        remaining_header: int,
        limit_header: int,
        date_header: datetime.datetime,
        reset_at_header: datetime.datetime,
    ) -> None:
        """Update the rate limits for a bucket using info from a response.

        Parameters
        ----------
        compiled_route : :obj:`hikari.net.routes.CompiledRoute`
            The compiled route to get the bucket for.
        bucket_header : :obj:`str`, optional
            The ``X-RateLimit-Bucket`` header that was provided in the response,
            or :obj:`None` if not present.
        remaining_header : :obj:`int`
            The ``X-RateLimit-Remaining`` header cast to an :obj:`int`.
        limit_header : :obj:`int`
            The ``X-RateLimit-Limit`` header cast to an :obj:`int`.
        date_header : :obj:`datetime.datetime`
            The ``Date`` header value as a :obj:`datetime.datetime`.
        reset_at_header : :obj:`datetime.datetime`
            The ``X-RateLimit-Reset`` header value as a
            :obj:`datetime.datetime`.
        """
        self.routes_to_hashes[compiled_route] = bucket_header

        real_bucket_hash = compiled_route.create_real_bucket_hash(bucket_header)

        if real_bucket_hash not in self.real_hashes_to_buckets:
            bucket = HTTPBucketRateLimiter(real_bucket_hash, compiled_route)
            self.real_hashes_to_buckets[real_bucket_hash] = bucket
        else:
            bucket = self.real_hashes_to_buckets[real_bucket_hash]

        reset_after = (reset_at_header - date_header).total_seconds()
        reset_at_monotonic = time.perf_counter() + reset_after
        bucket.update_rate_limit(remaining_header, limit_header, reset_at_monotonic)


class ExponentialBackOff:
    r"""Implementation of an asyncio-compatible exponential back-off algorithm with random jitter.

    .. math::

        t_{backoff} = b^{i} +  m \\cdot rand()

    Such that :math:`t_{backoff}` is the backoff time, :math:`b` is the base,
    :math:`i` is the increment that increases by 1 for each invocation, and
    :math:`m` is the jitter multiplier. :math:`rand()` returns a value in the
    range :math:`[0,1)`.

    Parameters
    ----------
    base : :obj:`float`
        The base to use. Defaults to ``2``.
    maximum : :obj:`float`, optional
        If not :obj:`None`, then this is the max value the backoff can be in a
        single iteration before an :obj:`asyncio.TimeoutError` is raised.
        Defaults to ``64`` seconds.
    jitter_multiplier : :obj:`float`
        The multiplier for the random jitter. Defaults to ``1``.
        Set to ``0`` to disable jitter.
    """

    __slots__ = ("base", "increment", "maximum", "jitter_multiplier")

    #: The base to use. Defaults to 2.
    #:
    #: :type: :obj:`float`
    base: typing.Final[float]

    #: The current increment.
    #:
    #: :type: :obj:`int`
    increment: int

    #: If not :obj:`None`, then this is the max value the backoff can be in a
    #: single iteration before an :obj:`asyncio.TimeoutError` is raised.
    #:
    #: :type: :obj:`float`, optional
    maximum: typing.Optional[float]

    #: The multiplier for the random jitter. Defaults to ``1``.
    #: Set to ``0`` to disable jitter.
    #:
    #: :type: :obj:`float`
    jitter_multiplier: typing.Final[float]

    def __init__(self, base: float = 2, maximum: typing.Optional[float] = 64, jitter_multiplier: float = 1) -> None:
        self.base = base
        self.maximum = maximum
        self.increment = 0
        self.jitter_multiplier = jitter_multiplier

    def __next__(self) -> float:
        """Get the next back off to sleep by."""
        value = self.base ** self.increment

        self.increment += 1

        if self.maximum is not None and value >= self.maximum:
            raise asyncio.TimeoutError()

        value += random.random() * self.jitter_multiplier  # nosec
        return value

    def __iter__(self) -> "ExponentialBackOff":
        """Return this object, as it is an iterator."""
        return self

    def reset(self) -> None:
        """Reset the exponential back-off."""
        self.increment = 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Utilities for creating and naming loggers with consistent names.

|internal|
"""

__all__ = ["get_named_logger"]

import logging
import typing


def get_named_logger(obj: typing.Any, *extra_objs: typing.Any) -> logging.Logger:
    """Build an appropriately named logger.

    If the passed object is an instance of a class, the class is used instead.

    If a class is provided/used, then the class name is used to name the logger.

    If a string is provided, then the string is used as the name. This is not recommended.

    Parameters
    ----------
    obj
        The object to study to produce a logger for.
    extra_objs
        optional extra components to add to the end of the logger name.

    Returns
    -------
    :obj:`logging.Logger`
        A created logger.
    """
    if not isinstance(obj, str):
        if not isinstance(obj, type):
            obj = type(obj)

        obj = obj.__qualname__

    if extra_objs:
        extras = ", ".join(map(str, extra_objs))
        obj = f"{obj}[{extras}]"

    return logging.getLogger(obj)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe Users on Discord."""
__all__ = ["User", "MyUser", "UserFlag", "PremiumType"]

import enum
import typing

import attr

from hikari import entities
from hikari import snowflakes
from hikari.internal import urls
from hikari.internal import marshaller


@enum.unique
class UserFlag(enum.IntFlag):
    """The known user flags that represent account badges."""

    #: None
    NONE = 0

    #: Discord Empoloyee
    DISCORD_EMPLOYEE = 1 << 0

    #: Discord Partner
    DISCORD_PARTNER = 1 << 1

    #: HypeSquad Events
    HYPESQUAD_EVENTS = 1 << 2

    #: Bug Hunter Level 1
    BUG_HUNTER_LEVEL_1 = 1 << 3

    #: House of Bravery
    HOUSE_BRAVERY = 1 << 6

    #: House of Brilliance
    HOUSE_BRILLIANCE = 1 << 7

    #: House of Balance
    HOUSE_BALANCE = 1 << 8

    #: Early Supporter
    EARLY_SUPPORTER = 1 << 9

    #: Team user
    TEAM_USER = 1 << 10

    #: System
    SYSTEM = 1 << 12

    #: Bug Hunter Level 2
    BUG_HUNTER_LEVEL_2 = 1 << 14

    #: Verified Bot
    VERIFIED_BOT = 1 << 16

    #: Verified Bot Developer
    VERIFIED_BOT_DEVELOPER = 1 << 17


@enum.unique
class PremiumType(enum.IntEnum):
    """The types of Nitro."""

    #: No premium.
    NONE = 0

    #: Premium including basic perks like animated emojis and avatars.
    NITRO_CLASSIC = 1

    #: Premium including all perks (e.g. 2 server boosts).
    NITRO = 2


@marshaller.marshallable()
@attr.s(slots=True)
class User(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents a user."""

    #: This user's discriminator.
    #:
    #: :type: :obj:`str`
    discriminator: str = marshaller.attrib(deserializer=str)

    #: This user's username.
    #:
    #: :type: :obj:`str`
    username: str = marshaller.attrib(deserializer=str)

    #: This user's avatar hash, if set.
    #:
    #: :type: :obj:`str`, optional
    avatar_hash: typing.Optional[str] = marshaller.attrib(raw_name="avatar", deserializer=str, if_none=None)

    #: Whether this user is a bot account.
    #:
    #: :type: :obj:`bool`
    is_bot: bool = marshaller.attrib(raw_name="bot", deserializer=bool, if_undefined=False)

    #: Whether this user is a system account.
    #:
    #: :type: :obj:`bool`
    is_system: bool = marshaller.attrib(raw_name="system", deserializer=bool, if_undefined=False)

    #: The public flags for this user.
    #:
    #: Note
    #: ----
    #: This will be :obj:`None` if it's a webhook user.
    #:
    #:
    #: :type: :obj:`UserFlag`, optional
    public_flags: typing.Optional[UserFlag] = marshaller.attrib(deserializer=UserFlag, if_undefined=None)

    @property
    def avatar_url(self) -> str:
        """URL for this user's custom avatar if set, else default."""
        return self.format_avatar_url()

    def format_avatar_url(self, fmt: typing.Optional[str] = None, size: int = 4096) -> str:
        """Generate the avatar URL for this user's custom avatar if set, else their default avatar.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this URL, defaults to ``png`` or ``gif``.
            Supports ``png``, ``jpeg``, ``jpg``, ``webp`` and ``gif`` (when
            animated). Will be ignored for default avatars which can only be
            ``png``.
        size : :obj:`int`
            The size to set for the URL, defaults to ``4096``.
            Can be any power of two between 16 and 4096.
            Will be ignored for default avatars.

        Returns
        -------
        :obj:`str`
            The string URL.

        Raises
        ------
        :obj:`ValueError`
            If ``size`` is not a power of two or not between 16 and 4096.
        """
        if not self.avatar_hash:
            return urls.generate_cdn_url("embed/avatars", str(self.default_avatar), fmt="png", size=None)
        if fmt is None and self.avatar_hash.startswith("a_"):
            fmt = "gif"
        elif fmt is None:
            fmt = "png"
        return urls.generate_cdn_url("avatars", str(self.id), self.avatar_hash, fmt=fmt, size=size)

    @property
    def default_avatar(self) -> int:
        """Integer representation of this user's default avatar."""
        return int(self.discriminator) % 5


@marshaller.marshallable()
@attr.s(slots=True)
class MyUser(User):
    """Represents a user with extended oauth2 information."""

    #: Whether the user's account has 2fa enabled.
    #: Requires the ``identify`` scope.
    #:
    #: :type: :obj:`bool`, optional
    is_mfa_enabled: typing.Optional[bool] = marshaller.attrib(
        raw_name="mfa_enabled", deserializer=bool, if_undefined=None
    )

    #: The user's set language, requires the ``identify`` scope.
    #:
    #: :type: :obj:`str`, optional
    locale: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: Whether the email for this user's account has been verified.
    #: Requires the ``email`` scope.
    #:
    #: :type: :obj:`bool`, optional
    is_verified: typing.Optional[bool] = marshaller.attrib(raw_name="verified", deserializer=bool, if_undefined=None)

    #: The user's set email, requires the ``email`` scope.
    #: This will always be :obj:`None` for bots.
    #:
    #: :type: :obj:`str`, optional
    email: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None)

    #: This user account's flags, requires the ``identify`` scope.
    #:
    #: :type: :obj:`UserFlag`, optional
    flags: typing.Optional[UserFlag] = marshaller.attrib(deserializer=UserFlag, if_undefined=None)

    #: The type of Nitro Subscription this user account had.
    #: Requires the ``identify`` scope and will always be :obj:`None` for bots.
    #:
    #: :type: :obj:`PremiumType`, optional
    premium_type: typing.Optional[PremiumType] = marshaller.attrib(deserializer=PremiumType, if_undefined=None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities related to discord's Oauth2 flow."""
__all__ = ["Application", "ApplicationOwner", "OwnGuild", "Team", "TeamMember", "TeamMembershipState"]

import enum
import typing

import attr

from hikari import entities
from hikari import guilds
from hikari import permissions
from hikari import snowflakes
from hikari import users
from hikari.internal import urls
from hikari.internal import marshaller


@enum.unique
class ConnectionVisibility(enum.IntEnum):
    """Describes who can see a connection with a third party account."""

    #: Only you can see the connection.
    NONE = 0

    #: Everyone can see the connection.
    EVERYONE = 1


@marshaller.marshallable()
@attr.s(slots=True)
class OwnConnection(entities.HikariEntity, entities.Deserializable):
    """Represents a user's connection with a third party account.

    Returned by the ``GET Current User Connections`` endpoint.
    """

    #: The string ID of the third party connected account.
    #:
    #: Warning
    #: -------
    #: Seeing as this is a third party ID, it will not be a snowflake.
    #:
    #:
    #: :type: :obj:`str`
    id: str = marshaller.attrib(deserializer=str)

    #: The username of the connected account.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: The type of service this connection is for.
    #:
    #: :type: :obj:`str`
    type: str = marshaller.attrib(deserializer=str)

    #: Whether the connection has been revoked.
    #:
    #: :type: :obj:`bool`
    is_revoked: bool = marshaller.attrib(raw_name="revoked", deserializer=bool, if_undefined=False)

    #: A sequence of the partial guild integration objects this connection has.
    #:
    #: :type: :obj:`typing.Sequence` [ :obj:`hikari.guilds.PartialGuildIntegration` ]
    integrations: typing.Sequence[guilds.PartialGuildIntegration] = marshaller.attrib(
        deserializer=lambda payload: [
            guilds.PartialGuildIntegration.deserialize(integration) for integration in payload
        ],
        if_undefined=list,
    )

    #: Whether the connection has been verified.
    #:
    #: :type: :obj:`bool`
    is_verified: bool = marshaller.attrib(raw_name="verified", deserializer=bool)

    #: Whether friends should be added based on this connection.
    #:
    #: :type: :obj:`bool`
    is_friend_syncing: bool = marshaller.attrib(raw_name="friend_sync", deserializer=bool)

    #: Whether activities related to this connection will be shown in the
    #: user's presence updates.
    #:
    #: :type: :obj:`bool`
    is_showing_activity: bool = marshaller.attrib(raw_name="show_activity", deserializer=bool)

    #: The visibility of the connection.
    #:
    #: :type: :obj:`ConnectionVisibility`
    visibility: ConnectionVisibility = marshaller.attrib(deserializer=ConnectionVisibility)


@marshaller.marshallable()
@attr.s(slots=True)
class OwnGuild(guilds.PartialGuild):
    """Represents a user bound partial guild object."""

    #: Whether the current user owns this guild.
    #:
    #: :type: :obj:`bool`
    is_owner: bool = marshaller.attrib(raw_name="owner", deserializer=bool)

    #: The guild level permissions that apply to the current user or bot.
    #:
    #: :type: :obj:`hikari.permissions.Permission`
    my_permissions: permissions.Permission = marshaller.attrib(
        raw_name="permissions", deserializer=permissions.Permission
    )


@enum.unique
class TeamMembershipState(enum.IntEnum):
    """Represents the state of a user's team membership."""

    #: Denotes the user has been invited to the team but has yet to accept.
    INVITED = 1

    #: Denotes the user has accepted the invite and is now a member.
    ACCEPTED = 2


@marshaller.marshallable()
@attr.s(slots=True)
class TeamMember(entities.HikariEntity, entities.Deserializable):
    """Represents a member of a Team."""

    #: The state of this user's membership.
    #:
    #: :type: :obj:`TeamMembershipState`
    membership_state: TeamMembershipState = marshaller.attrib(deserializer=TeamMembershipState)

    #: This member's permissions within a team.
    #: Will always be ``["*"]`` until Discord starts using this.
    #:
    #: :type: :obj:`typing.Set` [ :obj:`str` ]
    permissions: typing.Set[str] = marshaller.attrib(deserializer=set)

    #: The ID of the team this member belongs to.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`
    team_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The user object of this team member.
    #:
    #: :type: :obj:`hikari.users.User`
    user: users.User = marshaller.attrib(deserializer=users.User.deserialize)


@marshaller.marshallable()
@attr.s(slots=True)
class Team(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents a development team, along with all its members."""

    #: The hash of this team's icon, if set.
    #:
    #: :type: :obj:`str`, optional
    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str)

    #: The member's that belong to this team.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`hikari.snowflakes.Snowflake`, :obj:`TeamMember` ]
    members: typing.Mapping[snowflakes.Snowflake, TeamMember] = marshaller.attrib(
        deserializer=lambda members: {m.user.id: m for m in map(TeamMember.deserialize, members)}
    )

    #: The ID of this team's owner.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`
    owner_user_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    @property
    def icon_url(self) -> typing.Optional[str]:
        """URL of this team's icon, if set."""
        return self.format_icon_url()

    def format_icon_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the icon URL for this team if set.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this URL, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ``webp``.
        size : :obj:`int`
            The size to set for the URL, defaults to ``4096``. Can be any power
            of two between 16 and 4096 inclusive.

        Returns
        -------
        :obj:`str`, optional
            The string URL.

        Raises
        ------
        :obj:`ValueError`
            If ``size`` is not a power of two or not between 16 and 4096.
        """
        if self.icon_hash:
            return urls.generate_cdn_url("team-icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None


@marshaller.marshallable()
@attr.s(slots=True)
class ApplicationOwner(users.User):
    """Represents the user who owns an application, may be a team user."""

    #: This user's flags.
    #:
    #: :type: :obj:`hikari.users.UserFlag`
    flags: int = marshaller.attrib(deserializer=users.UserFlag)

    @property
    def is_team_user(self) -> bool:
        """If this user is a Team user (the owner of an application that's owned by a team)."""
        return bool((self.flags >> 10) & 1)


@marshaller.marshallable()
@attr.s(slots=True)
class Application(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents the information of an Oauth2 Application."""

    #: The name of this application.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: The description of this application, will be an empty string if unset.
    #:
    #: :type: :obj:`str`
    description: str = marshaller.attrib(deserializer=str)

    #: Whether the bot associated with this application is public.
    #: Will be :obj:`None` if this application doesn't have an associated bot.
    #:
    #: :type: :obj:`bool`, optional
    is_bot_public: typing.Optional[bool] = marshaller.attrib(
        raw_name="bot_public", deserializer=bool, if_undefined=None
    )

    #: Whether the bot associated with this application is requiring code grant
    #: for invites. Will be :obj:`None` if this application doesn't have a bot.
    #:
    #: :type: :obj:`bool`, optional
    is_bot_code_grant_required: typing.Optional[bool] = marshaller.attrib(
        raw_name="bot_require_code_grant", deserializer=bool, if_undefined=None
    )

    #: The object of this application's owner.
    #: This should always be :obj:`None` in application objects retrieved
    #: outside Discord's oauth2 flow.
    #:
    #: :type: :obj:`ApplicationOwner`, optional
    owner: typing.Optional[ApplicationOwner] = marshaller.attrib(
        deserializer=ApplicationOwner.deserialize, if_undefined=None
    )

    #: A collection of this application's rpc origin URLs, if rpc is enabled.
    #:
    #: :type: :obj:`typing.Set` [ :obj:`str` ], optional
    rpc_origins: typing.Optional[typing.Set[str]] = marshaller.attrib(deserializer=set, if_undefined=None)

    #: This summary for this application's primary SKU if it's sold on Discord.
    #: Will be an empty string if unset.
    #:
    #: :type: :obj:`str`
    summary: str = marshaller.attrib(deserializer=str)

    #: The base64 encoded key used for the GameSDK's ``GetTicket``.
    #:
    #: :type: :obj:`bytes`, optional
    verify_key: typing.Optional[bytes] = marshaller.attrib(
        deserializer=lambda key: bytes(key, "utf-8"), if_undefined=None
    )

    #: The hash of this application's icon, if set.
    #:
    #: :type: :obj:`str`, optional
    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str, if_undefined=None)

    #: This application's team if it belongs to one.
    #:
    #: :type: :obj:`Team`, optional
    team: typing.Optional[Team] = marshaller.attrib(deserializer=Team.deserialize, if_undefined=None, if_none=None)

    #: The ID of the guild this application is linked to
    #: if it's sold on Discord.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`, optional
    guild_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )

    #: The ID of the primary "Game SKU" of a game that's sold on Discord.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`, optional
    primary_sku_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )

    #: The URL slug that links to this application's store page
    #: if it's sold on Discord.
    #:
    #: :type: :obj:`str`, optional
    slug: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: The hash of this application's cover image on it's store, if set.
    #:
    #: :type: :obj:`str`, optional
    cover_image_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="cover_image", deserializer=str, if_undefined=None
    )

    @property
    def icon_url(self) -> typing.Optional[str]:
        """URL for this team's icon, if set."""
        return self.format_icon_url()

    def format_icon_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the icon URL for this application if set.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this URL, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ```webp``.
        size : :obj:`int`
            The size to set for the URL, defaults to ``4096``.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        :obj:`str`, optional
            The string URL.

        Raises
        ------
        :obj:`ValueError`
            If ``size`` is not a power of two or not between 16 and 4096.
        """
        if self.icon_hash:
            return urls.generate_cdn_url("app-icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None

    @property
    def cover_image_url(self) -> typing.Optional[str]:
        """URL for this icon's store cover image, if set."""
        return self.format_cover_image_url()

    def format_cover_image_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this application's store page's cover image is set and applicable.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this URL, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ``webp``.
        size : :obj:`int`
            The size to set for the URL, defaults to ``4096``.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        :obj:`str`, optional
            The string URL.

        Raises
        ------
        :obj:`ValueError`
            If ``size`` is not a power of two or not between 16 and 4096.
        """
        if self.cover_image_hash:
            return urls.generate_cdn_url("app-assets", str(self.id), self.cover_image_hash, fmt=fmt, size=size)
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Datastructure bases."""
__all__ = ["HikariEntity", "Serializable", "Deserializable", "RawEntityT", "UNSET"]

import abc
import typing

import attr

from hikari.internal import marshaller
from hikari.internal import meta

RawEntityT = typing.Union[
    None, bool, int, float, str, bytes, typing.Sequence[typing.Any], typing.Mapping[str, typing.Any]
]

T_contra = typing.TypeVar("T_contra", contravariant=True)
T_co = typing.TypeVar("T_co", covariant=True)


class Unset(metaclass=meta.SingletonMeta):
    """A singleton value that represents an unset field."""

    def __bool__(self):
        return False

    def __repr__(self):
        return type(self).__name__.upper()

    __str__ = __repr__


#: A variable used for certain update events where a field being unset will
#: mean that it's not being acted on, mostly just seen attached to event models.
UNSET = Unset()


@marshaller.marshallable()
@attr.s(slots=True)
class HikariEntity(metaclass=abc.ABCMeta):
    """The base for any entity used in this API."""


class Deserializable:
    """Mixin that enables the class to be deserialized from a raw entity."""

    __slots__ = ()

    @classmethod
    def deserialize(cls: typing.Type[T_contra], payload: RawEntityT) -> T_contra:
        """Deserialize the given payload into the object.

        Parameters
        ----------
        payload
            The payload to deserialize into the object.
        """
        return marshaller.HIKARI_ENTITY_MARSHALLER.deserialize(payload, cls)


class Serializable:
    """Mixin that enables an instance of the class to be serialized."""

    __slots__ = ()

    def serialize(self: T_co) -> RawEntityT:
        """Serialize this instance into a naive value."""
        return marshaller.HIKARI_ENTITY_MARSHALLER.serialize(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Defines a facade around :obj:`hikari.clients.shard_clients.ShardClient`.

This provides functionality such as keeping multiple shards alive
"""

__all__ = ["GatewayManager"]

import asyncio
import datetime
import math
import time
import typing

from hikari import events
from hikari import gateway_entities
from hikari import guilds
from hikari.clients import configs
from hikari.clients import runnable
from hikari.clients import shard_clients
from hikari.internal import conversions
from hikari.internal import more_logging
from hikari.state import event_dispatchers
from hikari.state import raw_event_consumers

ShardT = typing.TypeVar("ShardT", bound=shard_clients.ShardClient)


class GatewayManager(typing.Generic[ShardT], runnable.RunnableClient):
    """Provides a management layer for multiple-sharded bots.

    This also provides additional conduit used to connect up shards to the
    rest of this framework to enable management of dispatched events, etc.
    """

    def __init__(
        self,
        *,
        shard_ids: typing.Sequence[int],
        shard_count: int,
        config: configs.WebsocketConfig,
        url: str,
        raw_event_consumer_impl: raw_event_consumers.RawEventConsumer,
        shard_type: typing.Type[ShardT] = shard_clients.ShardClient,
        dispatcher: typing.Optional[event_dispatchers.EventDispatcher] = None,
    ) -> None:
        super().__init__(more_logging.get_named_logger(self, conversions.pluralize(shard_count, "shard")))
        self._is_running = False
        self.config = config
        self.raw_event_consumer = raw_event_consumer_impl
        self._dispatcher = dispatcher
        self.shards: typing.Dict[int, ShardT] = {
            shard_id: shard_type(shard_id, shard_count, config, raw_event_consumer_impl, url) for shard_id in shard_ids
        }
        self.shard_ids = shard_ids

    @property
    def latency(self) -> float:
        """Average heartbeat latency for all valid shards.

        This will return a mean of all the heartbeat intervals for all shards
        with a valid heartbeat latency that are in the
        :obj:`hikari.clients.shard_clients.ShardState.READY` state.

        If no shards are in this state, this will return ``float('nan')``
        instead.

        Returns
        -------
        :obj:`float`
            The mean latency for all ``READY`` shards that have sent at least
            one acknowledged ``HEARTBEAT`` payload. If there is not at least
            one shard that meets this criteria, this will instead return
            ``float('nan')``.
        """
        latencies = []
        for shard in self.shards.values():
            if not math.isnan(shard.latency):
                latencies.append(shard.latency)

        return sum(latencies) / len(latencies) if latencies else float("nan")

    async def start(self) -> None:
        """Start all shards.

        This safely starts all shards at the correct rate to prevent invalid
        session spam. This involves starting each shard sequentially with a
        5 second pause between each.
        """
        if self._is_running:
            raise RuntimeError("Cannot start a client twice.")

        if self._dispatcher is not None:
            await self._dispatcher.dispatch_event(events.StartingEvent())

        self._is_running = True
        self.logger.info("starting %s shard(s)", len(self.shards))
        start_time = time.perf_counter()
        for i, shard_id in enumerate(self.shard_ids):
            if i > 0:
                await asyncio.sleep(5)

            shard_obj = self.shards[shard_id]
            await shard_obj.start()
        finish_time = time.perf_counter()

        self.logger.info("started %s shard(s) in approx %.2fs", len(self.shards), finish_time - start_time)

        if self._dispatcher is not None:
            await self._dispatcher.dispatch_event(events.StartedEvent())

    async def join(self) -> None:
        """Wait for all shards to finish executing, then return."""
        await asyncio.gather(*(shard_obj.join() for shard_obj in self.shards.values()))

    async def close(self) -> None:
        """Close all shards.

        Waits for all shards to shut down before returning.
        """
        if self._is_running:
            self.logger.info("stopping %s shard(s)", len(self.shards))
            start_time = time.perf_counter()
            try:
                if self._dispatcher is not None:
                    await self._dispatcher.dispatch_event(events.StoppingEvent())
                await asyncio.gather(*(shard_obj.close() for shard_obj in self.shards.values()))
            finally:
                finish_time = time.perf_counter()
                self.logger.info("stopped %s shard(s) in approx %.2fs", len(self.shards), finish_time - start_time)
                self._is_running = False
                if self._dispatcher is not None:
                    await self._dispatcher.dispatch_event(events.StoppedEvent())

    async def update_presence(
        self,
        *,
        status: guilds.PresenceStatus = ...,
        activity: typing.Optional[gateway_entities.GatewayActivity] = ...,
        idle_since: typing.Optional[datetime.datetime] = ...,
        is_afk: bool = ...,
    ) -> None:
        """Update the presence of the user for all shards.

        This will only update arguments that you explicitly specify a value for.
        Any arguments that you do not explicitly provide some value for will
        not be changed.

        Warning
        -------
        This will only apply to connected shards.

        Notes
        -----
        If you wish to update a presence for a specific shard, you can do this
        by using the ``shards`` :obj:`typing.Mapping` to find the shard you
        wish to update.

        Parameters
        ----------
        status : :obj:`hikari.guilds.PresenceStatus`
            If specified, the new status to set.
        activity : :obj:`hikari.gateway_entities.GatewayActivity`, optional
            If specified, the new activity to set.
        idle_since : :obj:`datetime.datetime`, optional
            If specified, the time to show up as being idle since,
            or :obj:`None` if not applicable.
        is_afk : :obj:`bool`
            If specified, :obj:`True` if the user should be marked as AFK,
            or :obj:`False` otherwise.
        """
        await asyncio.gather(
            *(
                shard.update_presence(status=status, activity=activity, idle_since=idle_since, is_afk=is_afk)
                for shard in self.shards.values()
                if shard.connection_state
                in (shard_clients.ShardState.WAITING_FOR_READY, shard_clients.ShardState.READY)
            )
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Various functional types and metatypes.

|internal|
"""
__all__ = ["SingletonMeta", "Singleton"]


class SingletonMeta(type):
    """Metaclass that makes the class a singleton.

    Once an instance has been defined at runtime, it will exist until the
    interpreter that created it is terminated.

    Example
    -------
    .. code-block:: python

        >>> class Unknown(metaclass=SingletonMeta):
        ...     def __init__(self):
        ...         print("Initialized an Unknown!")

        >>> Unknown() is Unknown()    # True

    Note
    ----
    The constructors of instances of this metaclass must not take any arguments
    other than ``self``.

    Warning
    -------
    Constructing instances of class instances of this metaclass may not be
    thread safe.
    """

    ___instances___ = {}
    __slots__ = ()

    def __call__(cls):
        if cls not in SingletonMeta.___instances___:
            SingletonMeta.___instances___[cls] = super().__call__()
        return SingletonMeta.___instances___[cls]


class Singleton(metaclass=SingletonMeta):
    """Base type for anything implementing the :obj:`SingletonMeta` metaclass.

    Once an instance has been defined at runtime, it will exist until the
    interpreter that created it is terminated.

    Example
    -------
    .. code-block:: python

        >>> class MySingleton(Singleton):
        ...    pass

        >>> assert MySingleton() is MySingleton()

    Note
    ----
    The constructors of child classes must not take any arguments other than
    ``self``.

    Warning
    -------
    Constructing instances of this class or derived classes may not be thread
    safe.
    """
